from tests import unittest


from botocore.retries import quota


class TestRetryQuota(unittest.TestCase):
    def setUp(self):
        self.retry_quota = quota.RetryQuota(50)

    def test_can_acquire_amount(self):
        self.assertTrue(self.retry_quota.acquire(5))
        self.assertEqual(self.retry_quota.available_capacity, 45)

    def test_can_release_amount(self):
        self.assertTrue(self.retry_quota.acquire(5))
        self.assertEqual(self.retry_quota.available_capacity, 45)
        self.retry_quota.release(5)
        self.assertEqual(self.retry_quota.available_capacity, 50)

    def test_cant_exceed_max_capacity(self):
        self.assertTrue(self.retry_quota.acquire(5))
        self.assertEqual(self.retry_quota.available_capacity, 45)
        self.retry_quota.release(10)
        self.assertEqual(self.retry_quota.available_capacity, 50)

    def test_noop_if_at_max_capacity(self):
        self.retry_quota.release(10)
        self.assertEqual(self.retry_quota.available_capacity, 50)

    def test_cant_go_below_zero(self):
        self.assertTrue(self.retry_quota.acquire(49))
        self.assertEqual(self.retry_quota.available_capacity, 1)
        self.assertFalse(self.retry_quota.acquire(10))
        self.assertEqual(self.retry_quota.available_capacity, 1)



from botocore.stub import Stubber
from tests import BaseSessionTest


class TestSagemaker(BaseSessionTest):
    def setUp(self):
        super(TestSagemaker, self).setUp()
        self.region = 'us-west-2'
        self.client = self.session.create_client(
            'sagemaker', self.region)
        self.stubber = Stubber(self.client)
        self.stubber.activate()
        self.hook_calls = []

    def _hook(self, **kwargs):
        self.hook_calls.append(kwargs['event_name'])

    def tearDown(self):
        self.stubber.deactivate()

    def test_event_with_old_prefix(self):
        self.client.meta.events.register(
            'provide-client-params.sagemaker.ListEndpoints',
            self._hook
        )
        self.stubber.add_response('list_endpoints', {'Endpoints': []})
        self.client.list_endpoints()
        self.assertEqual(self.hook_calls, [
            'provide-client-params.sagemaker.ListEndpoints'
        ])

    def test_event_with_new_prefix(self):
        self.client.meta.events.register(
            'provide-client-params.api.sagemaker.ListEndpoints',
            self._hook
        )
        self.stubber.add_response('list_endpoints', {'Endpoints': []})
        self.client.list_endpoints()
        self.assertEqual(self.hook_calls, [
            'provide-client-params.sagemaker.ListEndpoints'
        ])



from tests import unittest


from botocore.retries import throttling


class TestCubicCalculator(unittest.TestCase):
    def create_cubic_calculator(self, starting_max_rate=10, beta=0.7,
                                scale_constant=0.4):
        return throttling.CubicCalculator(starting_max_rate=starting_max_rate,
                                          scale_constant=scale_constant,
                                          start_time=0, beta=beta)

    # For these tests, rather than duplicate the formulas in the tests,
    # I want to check against a fixed set of inputs with by-hand verified
    # values to ensure we're doing the calculations correctly.

    def test_starting_params(self):
        cubic = self.create_cubic_calculator(starting_max_rate=10)
        self.assertAlmostEqual(
            cubic.get_params_snapshot().k, 1.9574338205844317
        )

    def test_success_responses_until_max_hit(self):
        # For this test we're interested in the behavior less so than
        # the specific numbers.  There's a few cases we care about:
        #
        cubic = self.create_cubic_calculator(starting_max_rate=10)
        params = cubic.get_params_snapshot()
        start_k = params.k
        start_w_max = params.w_max
        # Before we get to t == start_k, our throttle is below our
        # max w_max
        assertLessEqual = self.assertLessEqual
        assertLessEqual(cubic.success_received(start_k / 3.0), start_w_max)
        assertLessEqual(cubic.success_received(start_k / 2.0), start_w_max)
        assertLessEqual(cubic.success_received(start_k / 1.1), start_w_max)
        # At t == start_k, we should be at w_max.
        self.assertAlmostEqual(cubic.success_received(timestamp=start_k), 10.0)
        # And once we pass start_k, we'll be above w_max.
        self.assertGreaterEqual(
            cubic.success_received(start_k * 1.1), start_w_max)
        self.assertGreaterEqual(
            cubic.success_received(start_k * 2.0), start_w_max)

    def test_error_response_decreases_rate_by_beta(self):
        # This is the default value here so we're just being explicit.
        cubic = self.create_cubic_calculator(starting_max_rate=10, beta=0.7)

        # So let's say we made it up to 8 TPS before we were throttled.
        rate_when_throttled = 8
        new_rate = cubic.error_received(current_rate=rate_when_throttled,
                                        timestamp=1)
        self.assertAlmostEqual(new_rate, rate_when_throttled * 0.7)

        new_params = cubic.get_params_snapshot()
        self.assertEqual(
            new_params,
            throttling.CubicParams(w_max=rate_when_throttled,
                                   k=1.8171205928321397,
                                   last_fail=1)
        )

    def test_t_0_should_match_beta_decrease(self):
        # So if I have beta of 0.6
        cubic = self.create_cubic_calculator(starting_max_rate=10, beta=0.6)
        # When I get throttled I should decrease my rate by 60%.
        new_rate = cubic.error_received(current_rate=10, timestamp=1)
        self.assertEqual(new_rate, 6.0)
        # And my starting rate at time t=1 should start at that new rate.
        self.assertAlmostEqual(cubic.success_received(timestamp=1), 6.0)



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from botocore.session import Session
from nose.tools import assert_false


def _all_services():
    session = Session()
    service_names = session.get_available_services()
    for service_name in service_names:
        yield session.get_service_model(service_name)


def _all_operations():
    for service_model in _all_services():
        for operation_name in service_model.operation_names:
            yield service_model.operation_model(operation_name)


def _assert_not_shadowed(key, shape):
    if not shape:
        return
    msg = (
        'Found shape "%s" that shadows the botocore response key "%s"'
    )
    assert_false(key in shape.members, msg % (shape.name, key))


def test_response_metadata_is_not_shadowed():
    for operation_model in _all_operations():
        shape = operation_model.output_shape
        yield _assert_not_shadowed, 'ResponseMetadata', shape


def test_exceptions_do_not_shadow():
    for service_model in _all_services():
        for shape in service_model.error_shapes:
            yield _assert_not_shadowed, 'ResponseMetadata', shape
            yield _assert_not_shadowed, 'Error', shape
